local jutils = require("utility.jutils")
local config = require("config")
local tiles = require("tiles")
local playerinventory = require("inventory.playerinventory")
local grid = require("utility.grid")
local input = require("utility.input")
local items = require("items")
local rendering = require("utility.rendering")

local physicalentity = require("entities.physicalentity")


local player = physicalentity:subclass("Player")

function player:init()
	physicalentity.init(self)

	self.texture = love.graphics.newImage("assets/entities/player.png")
	self.boundingbox = jutils.point.new(8, 12)
	self.textureorigin = jutils.point.new(8, 12)
	self.fallthrough = false
	self.grabrope = false
	self.freecamspeed = 4
	self.inventory = playerinventory:new()
	self.inventoryOpen = false
	self.walkspeed = 60
	self.acceleration = 800
	self.climbcooldown = 0
	self.showMouseTileDistance = -1
	self.hotbarslot = 1
	self.currentSelectedSlot = 0
	self.itemEquipped = nil
	self.lastItemEquipped = nil
	self.equipDebounce = false
	self.fast = false
	self.mouse = {
		leftDown = false, rightDown = false, 
		leftBounce = false, rightBounce = false,
		position = jutils.point.new(0, 0)
	}
	self.lclicked = false
	self.rclicked = false
	self.lusecooldown = 0
	self.rusecooldown = 0
	self.effects = {}
	self.escapeDebounce = false

	self.animation = {
		swingtimer = 0,
		swinging = false,
		item = nil,
	}

	self.inventory:addItem(items.HANDMADE_PICKAXE.id, 1)
	self.inventory:addItem(items.TNT_TILE.id, 999)
	self.inventory:addItem(items.ACTIVATOR_TILE.id, 999)
	self.inventory:addItem(items.ROPE_TILE.id, 999)

	input.mousepressed:connect(function(...)
		self:onMousePressed(...)
	end)
	input.mousereleased:connect(function(...)
		self:onMouseReleased(...)
	end)
	input.wheelmoved:connect(function(...)
		self:onWheelMoved(...)
	end)
	input.keypressed:connect(function(...)
		self:onKeyPressed(...)
	end)

	self:scroll(0)
end

function player:onMousePressed(x, y, button, istouch, presses)
	local result = self.inventory:clicked(x, y, button, istouch, presses)
	if result == false then
		if button == 1 then self.mouse.leftDown = true end
		if button == 2 then self.mouse.rightDown = true end
	end
end

function player:onMouseReleased(x, y, button, istouch, presses)
	if button == 1 then self.mouse.leftDown = false end
	if button == 2 then self.mouse.rightDown = false end
	--self:scroll(0)
end

function player:onKeyPressed(key)
	if key == "escape" then
		self.inventoryOpen = not self.inventoryOpen

		if self.inventoryOpen then
			self.inventory:fullOpen()
		else
			self.inventory:fullClose()
		end
	end

	-- item dropping, TODO: make it's own method
	if key == "q" then
		local itemdata = self.inventory:getHeldItem()
		if itemdata then
			local stack = self.itemEquippedStack
			if stack then
				local itemstack = self.world:addEntity("itemstack", stack[1], stack[2])

				-- TODO: add a timer to itemstacks
				-- to make them repel from player for a short time
				itemstack.nextposition = self.nextposition + jutils.point.new(-60, 0)

				stack[1] = 0
				stack[2] = 0
			end
		end
	end
end

function player:canLeftUse()
	if self.itemEquipped == nil then return false end
	if self.itemEquipped.leftuse == nil then return false end

	if self.lusecooldown <= 0 then
		if self.itemEquipped.repeating == true or self.mouse.leftBounce == false then
			return true
		end
	end
	return false
end

function player:canRightUse()
	if self.itemEquipped == nil then return false end
	if self.itemEquipped.rightuse == nil then return false end

	if self.rusecooldown <= 0 then
		if self.itemEquipped.repeating == true or self.mouse.rightBounce == false then
			return true
		end
	end
	return false
end

function player:leftuse()
	local item = self.itemEquipped
	self.lusecooldown = item.speed
	local success = item:leftuse(self)
	if success then
		self.animation.swingtimer = item.speed
		self.animation.swinging = true
		self.animation.item = self.itemEquipped
	end
end

function player:rightuse()
	local item = self.itemEquipped
	self.rusecooldown = item.speed
	item:rightuse(self)

	self.animation.swingtimer = item.speed
	self.animation.swinging = true
	self.animation.item = self.itemEquipped
end

function player:onWheelMoved(x, y)
	self:scroll(y)
end

function player:getMousePosition()
	return self.mouse.position
end

function player:scroll(change)
	self.hotbarslot = self.hotbarslot - change


	if self.hotbarslot < 1 then self.hotbarslot = 10 end
	if self.hotbarslot > 10 then self.hotbarslot = 1 end

	self.inventory.hotbarslot = self.hotbarslot
end

function player:collisionCallback(tileid, tilepos, separation, normal)

	if tiles:tileHasTag(tileid, "PlatformTile") then
		if normal.x == 0 and normal.y == -1 and self.fallthrough == false then

			local bottom = self.nextposition.y + self.boundingbox.y
			local realbottom = self.position.y + self.boundingbox.y

			local blockbottom = tilepos.y*config.TILE_SIZE+(config.TILE_SIZE/2)

			if realbottom <= blockbottom then
				self.nextposition = self.nextposition + separation
				self.falling = false
				self.velocity.y = 0
			end
		end
	end	

	if normal.y == 0 and normal.x ~= 0 then
		if self.velocity.y <= 0 and math.abs(self.velocity.x) > 10 then
			
			local tryY = tilepos.y*config.TILE_SIZE
			if tryY > self.nextposition.y and self.climbcooldown <= 0 then
				self.velocity.x = self.velocity.x * 0.6
				self.nextposition.y = (tryY-self.boundingbox.y)
				self.climbcooldown = 0.05
			end
		end
	end
end


function player:inventoryUpdate(dt)
	local itemid = self.inventory:getHeldItem()
	if itemid and itemid[1] > 0 then
		self.itemEquipped = items:getByID(itemid[1])
		self.itemEquippedStack = itemid
	else
		self.itemEquipped = nil
		self.itemEquippedStack = nil
	end

	if self.itemEquipped ~= self.lastItemEquipped then
		if self.lastItemEquipped and self.lastItemEquipped.unequip then
			self.lastItemEquipped:unequip(self)
		end
		if self.itemEquipped and self.itemEquipped.equip then
			self.itemEquipped:equip(self)
		end
		self.lastItemEquipped = self.itemEquipped
	end
end


function player:update(dt)

	---------------
	-- TODO: make statuseffect system that handles all this junk
	-- allows me to hook functions:
	-- effect:onComeUp(player), effect:onComeDown(player)

	self:inventoryUpdate(dt)

	self.animation.swingtimer = self.animation.swingtimer - dt

	if self.animation.swingtimer <= 0 then
		self.animation.swinging = false
	end
	
	self.boundingbox = jutils.point.new(8, 12)
	self.textureorigin = jutils.point.new(8, 12)
	self.scale.y = 1
	self.scale.x = 1

	-- status effects
	for index, effect in pairs(self.effects) do
		effect[3] = effect[3] - dt

		self.effects[index] = effect
		
		if effect[3] <= 0 then
			print(effect[1], "wore off")
			self.effects[index] = nil

		end

		if effect[1] == "MANLET" then
			self.boundingbox = jutils.point.new(6, 6)
			self.textureorigin = jutils.point.new(8, 12)
			self.scale.y = 0.5
			self.scale.x = 0.75
		end
	end

	physicalentity.update(self, dt)

	self.lusecooldown = self.lusecooldown - dt
	self.rusecooldown = self.rusecooldown - dt

	if self.mouse.leftDown then
		if self:canLeftUse() then
			self.mouse.leftBounce = true
			self:leftuse()
		else
			self.mouse.leftBounce = false
		end
	end

	if self.mouse.rightDown then
		if self:canRightUse() then
			self.mouse.rightBounce = true
			self:rightuse()
		else
			self.mouse.rightBounce = false
		end
	end

	self.climbcooldown = self.climbcooldown - dt
	self.velocity.x = jutils.math.clamp(-self.walkspeed, self.velocity.x, self.walkspeed)
	self.mouse.position = jutils.point.new(input.getTransformedMouse())

	local mm = (self.fast==true) and 3 or 1

	if self.frozen then
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_UP) then self.nextposition.y = self.nextposition.y - self.freecamspeed*mm end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_DOWN) then self.nextposition.y = self.nextposition.y + self.freecamspeed*mm end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_LEFT) then self.nextposition.x = self.nextposition.x - self.freecamspeed*mm end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_RIGHT) then self.nextposition.x = self.nextposition.x + self.freecamspeed*mm end
	else -- standard movement

		-- jumping
		if love.keyboard.isDown(config.keybinds.PLAYER_JUMP) and self.falling == false then 
			self.velocity.y = -150
			self.falling = true
		end

		-- rope and platform controls
		self.grabrope = love.keyboard.isDown(config.keybinds.PLAYER_MOVE_UP)
		self.fallthrough = love.keyboard.isDown(config.keybinds.PLAYER_MOVE_DOWN)

		-- left<->right movement
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_LEFT) then self.velocity.x = self.velocity.x - (self.acceleration*mm) * dt end
		if love.keyboard.isDown(config.keybinds.PLAYER_MOVE_RIGHT) then self.velocity.x = self.velocity.x + (self.acceleration*mm) * dt end

	end

	self.inventory:update(dt)
end

local function magnitude(x1, y1, x2, y2)
	return math.sqrt( (x2 - x1)^2 + (y2-y1)^2 )
end

function player:draw()
	
	physicalentity.draw(self)

	if self.animation.swinging then
		
		local data = self.animation.item
		local item = data.id
		if item>0 then

			local angle = ((self.animation.swingtimer/data.speed)*data.swingrange)-(data.swingrange/2)

			local dx, dy = math.cos(angle)-0.5, math.sin(angle)+0.5
			local r = 10
			rendering:drawItem(item, self.position.x-(dx*r), self.position.y-(dy*r), data.inWorldScale, -data.defaultRotation + angle, data.playerHoldPosition.x, data.playerHoldPosition.y)
		end
	end
	
	local mouse = self:getMousePosition()
	local tmx, tmy = grid.pixelToTileXY(mouse.x, mouse.y)
	local px, py = grid.pixelToTileXY(self.position.x, self.position.y)
	if magnitude(tmx, tmy, px, py) <= self.showMouseTileDistance then
		love.graphics.setColor(1, 1, 1, 0.5)
		love.graphics.setLineWidth(1)
		love.graphics.rectangle("line", tmx*config.TILE_SIZE, tmy*config.TILE_SIZE, config.TILE_SIZE, config.TILE_SIZE)
	end

	self.inventory:draw()
end

return player