local grid = require("utility.grid")
local input = require("utility.input")
local tiles = require("tiles")
local object = require("utility.object")
local jutils = require("utility.jutils")

local itemcollector = {}
local items = {}
local idinc = 1

love.graphics.setDefaultFilter("nearest", "nearest")

do
	function itemcollector:getByID(itemid)
		for name, data in pairs(items) do
			if data.id == itemid then
				return data
			end
		end
		error("No tile with ID".. itemid.. "exists!")
	end

	function itemcollector:getNumberOfItems()
		return idinc
	end

	local mt = {
		__index = function(t, k)
			if items[k] then return items[k] end
		end
	}
	setmetatable(itemcollector, mt)
end
---------------------------------------------------

local baseitem = object:subclass("Item")

function baseitem:filldefaults()
	self.name = "GENERIC_ITEM"
	self.displayname = "generic.Item"
	self.stack = 1
	self.repeating = true
	self.speed = 1
	self.rarity = 1
	self.id = idinc
	self.texture = love.graphics.newImage("assets/default.png")
	self.color = {1,1,1}
	self.animation = nil
	self.animationspeed = 0
	self.defaultRotation = 0
	self.inWorldScale = 2
	self.playerHoldPosition = jutils.point.new(4, 4)
	self.swingrange = math.pi/2

	self.equip = function(self, player) end
	self.unequip = function(self, player) end
	self.leftuse = function(self, player) end
	self.rightuse = function(self, player) end
end

function baseitem:init(name, data)
	self:filldefaults()
	self.name = name
	self.displayname = name

	local latestID = idinc
	if data.id~=nil then
		latestID = data.id
	end

	--print("regitem", name, latestID)

	idinc = latestID + 1

	-- override defaults
	for label, value in pairs(data) do
		if label == "texture" and type(value) == "string" then
			self["texture"] = love.graphics.newImage("assets/items/"..value)
		else
			self[label] = value
		end
	end

	items[name] = self
end

-------------------------------------

local function getMouseTile(world)
	local dx, dy = world:getMousePoint()
	local mx, my = grid.pixelToTileXY(dx, dy)
	return mx, my
end

local function getPlayerTile(playerentity)
	local pos = playerentity.position

	local tx, ty = grid.pixelToTileXY(pos.x, pos.y)
	return tx, ty
end

local function magnitude(x1, y1, x2, y2)
	return math.sqrt( (x2 - x1)^2 + (y2-y1)^2 )
end

local function pickaxeUse(pickaxe, player)
	local mx, my = input.getTransformedMouse()

	local tx, ty = grid.pixelToTileXY(mx, my)
	local px, py = getPlayerTile(player)

	if magnitude(px, py, tx, ty) < pickaxe.usedistance then

		local tileid = player.world:getTile(tx, ty)
		if tileid > 0 then
			
			player.world:damageTile(tx, ty, pickaxe.strength)
			return true
		end
	end
end

baseitem:new("MANLET_POTION", {
	displayname = "MANLET POTION",
	texture = "potion.png",
	color = {1, 0, 1},
	rarity = 2,
	tooltip = [[
Temporary shrinkage in a bottle.
Duration: 60 seconds		
	]],

	stack = 30,
	repeating = false,
	leftuse = function(self, player)
		
		local stack = player.itemEquippedStack

		stack[2] = stack[2]-1

		table.insert(player.effects, {"MANLET", 1, 60})

	end,
})

local PICKAXE_TOOLTIP = 
[[
{speed} Speed
{usedistance} Range
{strength} Pickaxe Power
]]

baseitem:new("HANDMADE_AXE", {
	displayname = "HANDMADE AXE",
	texture = "pickaxenew.png",
	speed = 1/4,
	tooltip = "Crude yet effective handtool.\n"..PICKAXE_TOOLTIP,
	stack = 1,
	usedistance = 6,
	strength = 1,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

})

baseitem:new("HANDMADE_PICKAXE", {
	displayname = "HANDMADE PICKAXE",
	texture = "pickaxenew.png",
	stack = 1,
	tooltip = "A rock and a stick binded with rope.\n"..PICKAXE_TOOLTIP,
	color = {1, 0.7, 0.7},
	speed = 1/4,
	usedistance = 6,
	swingrange = math.pi,
	playerHoldPosition = jutils.point.new(0, 8),
	strength = 1,
	defaultRotation = math.rad(45),
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

	leftuse = pickaxeUse,
})



baseitem:new("IRON_PICKAXE", {
	texture = "pickaxenew.png",
	color = {0.9, 0.75, 0.75},
	speed = 1/6,
	stack = 1,
	rarity = 2,
	usedistance = 8,
	strength = 2.5,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

	leftuse = pickaxeUse,
})



baseitem:new("COPPER_PICKAXE", {
	texture = "pickaxenew.png",
	color = {0.9, 0.45, 0.1},
	speed = 1/10,
	stack = 1,
	usedistance = 10,
	strength = 1,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = pickaxeUse,
})

baseitem:new("DOOR", {
	texture = "door.png",
	color = {0.6, 0.4, 0.1},
	stack = 99,
	usedistance = 8,
	speed = 1/10,
	leftuse = function(self, player)

		local mx, my = input.getTransformedMouse()
		local stack = player.itemEquippedStack
		local tx, ty = grid.pixelToTileXY(mx, my)

		local function isempty(world, x, y)
			if world:getTile(x, y) == 0 then return true end
		end
		

		if isempty(player.world, tx, ty) and isempty(player.world, tx, ty-1) and (not isempty(player.world, tx, ty+1)) then
			player.world:setTile(tx, ty, tiles.DOOR_BOTTOM.id)
			player.world:setTile(tx, ty-1, tiles.DOOR_TOP.id)
			stack[2] = stack[2] - 1
		end

		
	end,
})

baseitem:new("LEAD_PICKAXE", {
	texture = "pickaxenew.png",
	color = {0.35, 0.35, 0.45},
	speed = 1/4,
	stack = 1,
	usedistance = 12,
	strength = 4,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = pickaxeUse,
})

baseitem:new("IRON_INGOT", {
	displayname = "IRON INGOT",
	texture = "ingot.png",
	color = {1, 0.8, 0.8}
})

baseitem:new("COAL", {

})

baseitem:new("PAPER", {

})

baseitem:new("DEBUG_TOOL", {
	displayname = "DEBUG TOOL",
	speed = 1/10,
	stack = 1,
	repeating = false,
	rarity = -1,
	usedistance = 50,
	equip = function(self, player)
		player.showMouseTileDistance = self.usedistance
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,

	leftuse = function(self, player)
		local mx, my = input.getTransformedMouse()

		local tx, ty = grid.pixelToTileXY(mx, my)

		local state = player.world:getTileState(tx, ty)

		state = state + 1

		player.world:setTileState(tx, ty, state)

		player.world:tileUpdate(tx, ty)
		player.world:randomTick(tx, ty)

		print("SETSTATE "..tx..", "..ty..": "..state)
	end,
	rightuse = function(self, player)
		local mx, my = input.getTransformedMouse()

		local tx, ty = grid.pixelToTileXY(mx, my)

		print("GETSTATE "..tx..", "..ty..": "..player.world:getTileState(tx, ty))
	end,
})

baseitem:new("FERTILIZER", {
	displayname = "Fertilizer",
	speed = 1/10,
	repeating = true,
	specialdata = {usedistance=10},
	equip = function(self, player)
		player.showMouseTileDistance = 10
	end,
	unequip = function(self, player)
		player.showMouseTileDistance = -1
	end,
	leftuse = function(self, player)
		local mx, my = input.getTransformedMouse()

		local tx, ty = grid.pixelToTileXY(mx, my)

	end,
})


local multitile = baseitem:subclass("MultiTileItem")

function multitile:init()

end


-- Add all tiles to the items list
local num = tiles:getNumberOfTiles()
for i = 1, num-1 do
	local data = tiles:getByID(i)
	baseitem:new(data.name.."_TILE", {
		displayname = string.gsub(data.name, "_", " "),
		texture = data.texture,
		color = data.color,
		speed = 1/6,
		animation = data.animation,
		animationspeed = data.animationspeed,
		tileid = data.id,
		usedistance = 8,
		inWorldScale = 1,
		
		stack = 999,
		equip = function(self, player)
			player.showMouseTileDistance = self.usedistance
		end,

		unequip = function(self, player)
			player.showMouseTileDistance = -1
		end,
		leftuse = function(self, player)
			local itemdata = player.itemEquipped
			local tiledata = tiles:getByID(self.tileid)


			local stack = player.itemEquippedStack
			local mx, my = input.getTransformedMouse()

			local tx, ty = grid.pixelToTileXY(mx, my)
			local px, py = getPlayerTile(player)
			local canPlace = true
			if magnitude(px, py, tx, ty) > self.usedistance then
				canPlace = false
			end

			if player.world:getTile(tx, ty) ~= 0 then
				canPlace = false
			end

			if tiledata.validplacement then
				if not tiledata.validplacement(player.world, tx, ty) then
					canPlace = false
				end
			end

			if canPlace then
				player.world:setTile(tx, ty, self.tileid)
				stack[2] = stack[2] - 1
				return true
			end
			return false
		end,
	})
end



return itemcollector