local jutils = require("utility.jutils")
local config = require("config")

local tilecollector = {}
local tiles = {}
local idinc = 0

local defaultQuad = love.graphics.newQuad(0, 0, 8, 8, 8, 8)

do

	function tilecollector:tileHasTag(tileid, looktag)

		local tiledata

		if type(tileid) == "number" then
			tiledata = tilecollector:getByID(tileid)
		else
			tiledata = tileid
		end


		for _, tag in pairs(tiledata.tags) do
			if tag == looktag then
				return true
			end
		end
		return false
	end

	function tilecollector:getByID(tileid)
		assert(type(tileid=="number"), "TileID must be a number!")
		for name, data in pairs(tiles) do
			if data.id == tileid then
				return data
			end
		end
		error("No tile with ID".. tileid.. "exists!", 2)
	end

	function tilecollector:getNumberOfTiles()
		return idinc
	end

	local mt = {
		__index = function(t, k)

			if tiles[k] then return tiles[k] end
		end
	}

	setmetatable(tilecollector, mt)
end
----------------------------------

love.graphics.setDefaultFilter("nearest", "nearest")
---------------------------------------------------
-- tile update and random tile update callbacks

local function isSolid(tileid)
	local data = tilecollector:getByID(tileid)
	return data.solid
end

local function aboveSolidTile(world, tileid, x, y)
	if not isSolid(world:getTile(x, y+1)) then
		world:setTile(x, y, tiles.AIR.id)
	end
end

local function adjacentToNonSolidTile(world, x, y)
	for dx = -1, 1 do
		if not isSolid(world:getTile(x+dx, y)) then return true end	
	end
	for dy = -1, 1 do
		if not isSolid(world:getTile(x, y+dy)) then return true end
	end
	return false
end

local function grassCheck(world, x, y)

	local stayalive = adjacentToNonSolidTile(world, x, y)

	if not stayalive then world:setTile(x, y, tiles.DIRT.id) return end

	-- spreading onto dirt
	for dx = -1, 1 do
		for dy = -1, 1 do
			if world:getTile(x+dx, y+dy) == tiles.DIRT.id and adjacentToNonSolidTile(world, x+dx, y+dy) then
				world:setTile(x+dx, y+dy, tiles.GRASS.id)
			end
		end
	end
end

local function logValidCheck(world, x, y)
	if not isSolid(world:getTile(x, y+1)) then
		world:setTile(x, y, tiles.AIR.id, true)
	end
end





local function leavesRandomUpdate(world, x, y)
	local die = true

	for dx = -2, 2 do
		for dy = -2, 2 do
			if world:getTile(x+dx, y+dy) == tiles.LOG.id then
				die = false
			end
		end
	end

	if die then
		world:setTile(x, y, tiles.AIR.id)
	end
end

local function bluegrassCheck(world, x, y)
	local stayalive = adjacentToNonSolidTile(world, x, y)

	if not stayalive then world:setTile(x, y, tiles.MUD.id) return end

	-- spreading onto dirt
	for dx = -1, 1 do
		for dy = -1, 1 do
			if world:getTile(x+dx, y+dy) == tiles.MUD.id and adjacentToNonSolidTile(world, x+dx, y+dy) then
				world:setTile(x+dx, y+dy, tiles.BLUE_GRASS.id)
			end
		end
	end
end

local function explodeTNT(world, x, y, stk)
	stk = stk and stk+1 or 1
	
	local tntradius = 8
	local exp = world:addEntity("explosion", tntradius, tntradius+2, true)
	
	exp.position = jutils.point.new(x*config.TILE_SIZE, y*config.TILE_SIZE)
	if stk > 2 then return end
	for dx = -tntradius, tntradius do
		for dy = -tntradius, tntradius do
			if world:getTile(x+dx, y+dy) == tiles.TNT.id then
				explodeTNT(world, x+dx, y+dy, stk)
			end
		end
	end
end

local function activationcheck(world, x, y)
	for dx = -1, 1 do
		if world:getTile(x+dx, y) == tiles.TNT.id then
			world:setTile(x+dx, y, tiles.AIR.id, false)
			world:setTile(x, y, tiles.AIR.id, false)
			explodeTNT(world, x+dx, y)
			return
		end
	end
	for dy = -1, 1 do
		if world:getTile(x, y+dy) == tiles.TNT.id then
			world:setTile(x, y+dy, tiles.AIR.id, false)
			world:setTile(x, y, tiles.AIR.id, false)
			explodeTNT(world, x, y+dy)
			return
		end
	end
	
end





---------------------------------------------------

-- Tile constructors
local function tile(name, data)

	local default = {
		name = name,
		id = idinc,
		color = {1, 1, 1},
		drop = name.."_TILE",
		texture = love.graphics.newImage("assets/default.png"),
		tags = {},
		quad = nil,
		collide = true,
		solid = true,
		animation = nil,
		animationspeed = 4, -- animation frames/sec
		tileupdate = nil,
		randomupdate = nil,
		validplacement = nil,
		onplace = nil,
		onbreak = nil,
		customrender = nil,
		hardness = 2,
		light = 0,
	}

	local latestID = idinc
	if data.id~=nil then latestID = data.id end

	tiles[name] = {}
	for label, value in pairs(default) do
		tiles[name][label] = value
	end
	-- override default values
	for label, value in pairs(data) do
		if label == "texture" then
			tiles[name]["texture"] = love.graphics.newImage("assets/tiles/"..value)
		else
			tiles[name][label] = value
		end
	end

	idinc = latestID + 1
end
----------------------------------------------------
local function stoneTile(variantname, color)
	tile(variantname, {
		color = color,
		texture = "stone.png",
		hardness = 3,
	})
end

local function brickTile(variantname, color)
	tile(variantname, {
		color = color,
		texture = "brick.png",
		hardness = 4,
	})
end

-- Actual tiles begin here
-----------------------------------------------------------------------------

tile("VOID", {color = {0, 0, 0, 1}, texture = "blank.png", id = -1, solid = false, collide = false})
tile("AIR", {color = {0, 0, 0, 1}, texture = "blank.png", id = 0, solid = false, collide = false})

tile("GRASS", {
	texture = "soil.png",
	color = {0.2, 0.8, 0.2},
	randomupdate = grassCheck,
	tags = {"plantable-on"},
	drop = "DIRT_TILE",
})

tile("DIRT", {
	texture = "soil.png",
	color = {0.45, 0.25, 0.1},
	tags = {"plantable-on"},
	hardness = 1,
})

tile("BLUE_GRASS", {
	texture = "soil.png",
	color = {0.2, 0.2, 0.9},
	randomupdate = bluegrassCheck,
	tags = {"plantable-on"},
	drop = "DIRT_TILE",
})

tile("MUD", {
	texture = "soil.png",
	color = {0.35, 0.15, 0.05},
	tags = {"plantable-on"},
	hardness = 1,
})

tile("ACTIVATOR", {
	tileupdate = activationcheck
})

tile("TNT", {
	texture = "tnt.png",
	color = {1, 1, 1},
	hardness = 1,
})

local function torchValid(world, x, y)
	if not isSolid(world:getTile(x, y+1)) then
		return false
	end
	return true
end

tile("TORCH", {
	texture = "torch.png",
	color = {0.9, 0.9, 0.2},
	solid = false,
	collide = false,
	animation = {
		[1] = love.graphics.newQuad(0,0,8,8,32,8),
		[2] = love.graphics.newQuad(8,0,8,8,32,8),
		[3] = love.graphics.newQuad(16,0,8,8,32,8),
		[4] = love.graphics.newQuad(24,0,8,8,32,8),
	},
	light = 2,
	tileupdate = function(world, x, y)
		if not torchValid(world, x, y) then
			world:setTile(x, y, tiles.AIR.id, true)
		end
	end,
	validplacement = torchValid,
	hardness = 1,
})

local function ropeValid(world, x, y)
	if not (isSolid(world:getTile(x, y-1)) or world:getTile(x, y-1)==tiles.ROPE.id) then
		return false
	end
	return true
end

tile("ROPE", {
	texture = "rope.png",
	solid = false,
	collide = true,
	tileupdate = function(world, x, y)
		if not ropeValid(world, x, y) then
			world:setTile(x, y, tiles.AIR.id, true)
		end
	end,
	validplacement = ropeValid,
	hardness = 1,
})
stoneTile("STONE",         {  1,   1,    1})
stoneTile("STONE_RED",     {  1, 0.8,  0.8})
stoneTile("STONE_YELLOW",  {  1,   1, 0.85})
stoneTile("STONE_BROWN",   {0.9, 0.9, 0.75})

brickTile("GRAY_BRICK",   {0.6, 0.6, 0.6})
brickTile("YELLOW_BRICK", {0.9, 0.9, 0.5})
brickTile("RED_BRICK",    {0.9, 0.5, 0.5})
brickTile("DARK_BRICK",   {0.2, 0.2, 0.2})
brickTile("WHITE_BRICK",  {  1,   1,   1})

tile("LOG", {
	texture = "log.png",
	color = {0.6, 0.4, 0.1},
	solid = true,
	collide = false,
	tileupdate = logValidCheck,
	drop = "PLANK_TILE",
	hardness = 8,
	tags = {"LOG"}
})

tile("LEAVES", {
	texture = "leaves.png",
	color = {0.2, 0.8, 0.2},
	solid = false,
	collide = false,
	randomupdate = leavesRandomUpdate,
	hardness = 1,
})
-- IDEA: dead leaves

tile("PLANK", {
	texture = "planks.png",
	color = {0.7, 0.5, 0.2},
	hardness = 3,
})

tile("PLATFORM", {
	texture = "platform.png",
	color = {0.7, 0.5, 0.2},
	solid = false,
	collide = true,
	tags = {"PlatformTile"}
})

local function saplingValid(world, x, y)

end

local function plantValidityCheck(world, x, y)
	if tilecollector:tileHasTag(world:getTile(x, y+1), "plantable-on") then
		return true
		
	end
	return false
end

local function saplinggrow(world, x, y)
	
	local maketree = require("world.structures.maketree")

	local state = world:getTileState(x, y)+1

	if state > 8 then
		maketree(world, x, y+1)
		
	else
		world:setTileState(x, y, state)
	end
end

tile("SAPLING", {
	texture = "sapling.png",
	color = {1,1,1},
	solid = false,
	collide = false,
	tileupdate = function(world, x, y)
		if not plantValidityCheck then
			world:setTile(x, y, tiles.AIR.id, true)
		end
	end,
	randomupdate = saplinggrow,
	validplacement = plantValidityCheck,
	tags = {"PLANT"},
	hardness = 1,
})

local function tileArray(name, data, width, height)
	for x = 1, width do
		for y = 1, height do

			local finaldata = data



			tile(name.."_"..x.."_"..y, finaldata)
		end
	end
end

tile("DOOR_TOP", {
	texture = "door.png",
	color = {0.7, 0.5, 0.2},
	hardness = 3,
	quad = love.graphics.newQuad(0, 0, 8, 8, 8, 8),
	tileupdate = function(world, x, y)
		if world:getTile(x, y+1) ~= tiles.DOOR_BOTTOM.id then
			world:setTile(x, y, 0)
		end
	end,
	drop = "DOOR",
})

tile("DOOR_BOTTOM", {
	texture = "door.png",
	color = {0.7, 0.5, 0.2},
	hardness = 3,
	quad = love.graphics.newQuad(0, 8, 8, 8, 8, 8),
	tileupdate = function(world, x, y)
		if world:getTile(x, y-1) ~= tiles.DOOR_TOP.id then
			world:setTile(x, y, 0)
		end
	end,
	drop = "DOOR",
})

tile("IRON_ORE", {
	texture = "ore.png",
	color = {1, 0.8, 0.8}
})

tile("COPPER_ORE", {
	texture = "ore.png",
	color = {1, 0.45, 0}
})

tile("LEAD_ORE", {
	texture = "ore.png",
	color = {0.35, 0.35, 0.45}
})

tile("SILVER_ORE", {
	texture = "ore.png",
	color = {1,1,1}
})

tile("PALLADIUM_ORE", {
	texture = "ore.png",
	color = {0.9, 0.5, 0.9},
})

tile("CHROMIUM_ORE", {
	texture = "ore.png",
	color = {0.5, 1, 1},
})

tile("GOLD_ORE", {
	texture = "ore.png",
	color = {1, 1, 0.5}
})

tile("TIN_ORE", {
	texture = "ore.png",
	color = {0.6, 0.4, 0.4}
})

tile("ALUMINIUM_ORE", {
	texture = "ore.png",
	color = {0.9, 0.9, 0.9}
})

tile("NICKEL_ORE", {
	texture = "ore.png",
	color = {1, 0.5, 0.5},
})

tile("VANADIUM_ORE", {
	texture = "ore.png",
	color = {0.5, 1, 0.8},
})

tile("COBALT_ORE", {
	texture = "ore.png",
	color = {0.3, 0.3, 1}
})

tile("TITANIUM_ORE", {
	texture = "ore.png",
	color = {0.6, 0.6, 0.6},
})

tile("URANIUM_ORE", {
	texture = "ore.png",
	color = {0.6, 1, 0.6}
})


tile("SAND", {
	texture = "soil.png",
	color = {0.7, 0.7, 0.3},
	hardness = 1,
})

tile("SMALL_BRICK", {
	texture = "bricksmall.png",
	color = {0.7, 0.7, 0.7}
})

tile("SLANTED_BRICK", {
	texture = "brickslanted.png",
	color = {1, 1, 1}
})

local function getWaterLevel(world, x, y)
	local id = world:getTile(x, y)

	if id == tiles.AIR.id then
		return 0
	end

	if id == tiles.WATER.id then
		return world:getTileState(x, y)
	end

	return -1
end

local function setWaterLevel(world, x, y, state)
	local id = world:getTile(x, y)

	if id == tiles.WATER.id then
		world:setTileState(x, y, state)
	end

	if id == tiles.AIR.id then
		world:setTile(x, y, tiles.WATER.id, false, true)
		world:setTileState(x, y, state)
	end
end

local waterLevels = 8

tile("WATER",{
	color = {0, 0, 0.5, 0.5},
	texture = "blank.png",
	solid = false,
	collide = true,
	onplace = function(world, x, y)
		world:setTileState(x, y, waterLevels)
		print("water placed")
	end,
	customCollision = function(entity, separation, normal)
		entity.velocity.x = jutils.math.clamp(-30, entity.velocity.x, 30)
		entity.velocity.y = jutils.math.clamp(-150, entity.velocity.y, 40)
	end,
	customrender = function(state, dmg, light, tx, ty)
		love.graphics.setColor(0, 0, 0.75, 0.5)
		love.graphics.rectangle("fill",
			(tx) * config.TILE_SIZE,
			(ty) * config.TILE_SIZE + (config.TILE_SIZE) - (state),
			config.TILE_SIZE,
			(config.TILE_SIZE / 8)*state)
	end,
	tileupdate = function(world, x, y)
		local waterMax = 8

		local this = getWaterLevel(world, x, y)
		if this < 1 then
			world:setTile(x, y, tiles.AIR.id)
			return
		end
		
		-- try to flow downward
		local below = getWaterLevel(world, x, y+1)
		local function check(val)
			if val ~= -1 and val < waterMax then return true end
		end

		if check(below) and this > 0 then
			this = this-1
			below = below+1
			setWaterLevel(world, x, y, this)
			setWaterLevel(world, x, y+1, below)
			return
		end

		-- try to flow outwards once you're on the ground?
		local left = getWaterLevel(world, x-1, y)
		if check(left) and this > 0 then
			if this > left+1 then
				this = this-1
				left = left+1
			elseif this-left == 1 then
				local rand = math.random()
				if rand > 0.5 then
					this = this
					left = this
				else
					this = left
					left = left
				end
			end
			setWaterLevel(world, x, y, this)
			setWaterLevel(world, x-1, y, left)
		end

		local right = getWaterLevel(world, x+1, y)
		if check(right) and this > 0 then
			if this > right+1 then
				this = this-1
				right = right+1
			elseif this-right == 1 then
				local rand = math.random()
				if rand > 0.5 then
					this = this
					right = this
				else
					this = right
					right = right
				end
			end
			setWaterLevel(world, x, y, this)
			setWaterLevel(world, x+1, y, right)
		end
	end,
})


-- End tile construction
----------------------------------
return tilecollector