--[[
	This is the world class for cave game,
	likely to be the largest and most complex script so far.


	Method API:

	world:setFocusPoint(jutils.point fp)

	world:getTile(int tx, int ty) -> int tileid
	world:setTile(int tx, int ty, int tileid)

	-= Internal Methods =-
	world:getChunk(int cx, int cy) -> Chunk c
	world:unloadAllChunks()
	world:setchunkloaded(int cx, int cy)
	world:tileUpdate(int tx, int ty)

]]

local config = require("config")
local tiles = require("tiles")
local collision = require("utility.collision")
local jutils = require("utility.jutils")
local rendering = require("utility.rendering")
local items = require("items")

local grid = require("utility.grid")

----------------------------
-- World Class starts here

local world = {}

world.__index = world

function world.__init()
	local self = setmetatable({}, world)

	return self
end

function world.new(worldname)
	local self = world.__init()

	self.entitylist = {
		player = require("entities.player"),
		itemstack = require("entities.itemstack"),
		explosion = require("entities.explosion")
	}

	self.worldname = worldname
	self.entities = {}
	
	self.chunks = {}
	self.loaded = {}
	self.firstPassComplete = {}
	self.secondPassComplete = {}
	self.fullSimulation = {}
	self.seed = 0
	self.chunkdrawradius = config.CHUNK_DRAW_RADIUS
	self.generationradius = 1
	self.focuspoint = jutils.point.new(0, 0)
	self.terrainGenerator = require("world.generators.terraingenerator")
	self.structureGenerator = require("world.generators.structuregenerator")
	self.chunknum = 0
	self.firstpassnum = 0
	self.secondpassnum = 0
	self.loadednum = 0
	self.fullnum = 0
	self.dx = 0
	self.dy = 0
	self.ambientlight = 0
	return self
end

--[[
	HERE are the functions of note for scripts accessing a world instance:
]]

function world:addEntity(entityname, ...)
	local entity = self.entitylist[entityname]:new(...)
	entity.world = self
	table.insert(self.entities, entity)
	return entity
end
function world:getMousePoint()
	return self.dx, self.dy
end

function world:setFocusPoint(point)
	self.focuspoint = point
end

function world:iteratechunk(chunk)
	local function iter(chunk, i)
		
	end
end


function world:getTileDamage(tilex, tiley)
	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)

	if chunk then
		return chunk.damage[lx][ly]
	end
	return -1
end

function world:setTileDamage(tilex, tiley, damage)
	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)

	if chunk then
		return chunk.damage[lx][ly]
	end
	return -1
end

function world:damageTile(tilex, tiley, additionaldamage)
	assert(type(tilex) == "number")
	assert(type(tiley) == "number")
	assert(type(additionaldamage) == "number")

	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)
	if chunk then

		chunk.damage[lx][ly] = chunk.damage[lx][ly] + additionaldamage

		local tile = chunk.tiles[lx][ly]
		local maxdamage = tiles:getByID(tile).hardness

		if chunk.damage[lx][ly] >= maxdamage then
			self:setTile(tilex, tiley, 0, true)
		end
	else
		print("attempted to set damage in empty chunk at worldpos " .. tilex .. ", " .. tiley)
	end
end

-- returns the tileID at the given position
function world:getTile(tilex, tiley)
	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)

	if chunk then
		return chunk.tiles[lx][ly]
	end
	return -1
end

function world:getTileState(tilex, tiley)
	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)

	if chunk then
		return chunk.states[lx][ly]
	end
	return -1
end

function world:setTileState(tilex, tiley, state)
	assert(type(tilex) == "number")
	assert(type(tiley) == "number")
	assert(type(state) == "number")

	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)
	if chunk then
		if chunk.states[lx][ly] ~= state then


			chunk.states[lx][ly] = state

			self:tileUpdate(tilex, tiley - 1)
			self:tileUpdate(tilex, tiley + 1)
			self:tileUpdate(tilex + 1, tiley)
			self:tileUpdate(tilex - 1, tiley)
			self:tileUpdate(tilex, tiley)
		end
		
	else
		print("attempted to set tilestate in empty chunk:" .. state .. " at worldpos " .. tilex .. ", " .. tiley)
	end
end

-- sets the tile and fills the tileupdate flags of adjacent tiles
function world:setTile(tilex, tiley, tile, drop, ignorecallback)
	assert(type(tilex) == "number")
	assert(type(tiley) == "number")
	assert(type(tile) == "number")

	local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
	local chunk = self:getchunk(cx, cy)
	if chunk then
		if chunk.tiles[lx][ly] ~= tile then
			--self:setTileState(tilex, tiley, 0)

			chunk.states[lx][ly] = 0
			chunk.damage[lx][ly] = 0

			if drop then
				local oldid = chunk.tiles[lx][ly]
				local tiledata = tiles:getByID(oldid)
				if tiledata.drop ~= false then
					local item = items[tiledata.drop]

					local e = self:addEntity("itemstack", item.id, 1)
					e.nextposition.x = tilex*config.TILE_SIZE
					e.nextposition.y = tiley*config.TILE_SIZE
				end
			end

			chunk.tiles[lx][ly] = tile

			local tiledata = tiles:getByID(tile)
			if not ignorecallback then
				if tiledata.onplace then
					tiledata.onplace(self, tilex, tiley)
				end
			end
			

			self:tileUpdate(tilex, tiley)
			self:tileUpdate(tilex, tiley - 1)
			self:tileUpdate(tilex, tiley + 1)
			self:tileUpdate(tilex + 1, tiley)
			self:tileUpdate(tilex - 1, tiley)
		end
	else
		print("attempted to place in empty chunk:" .. tile .. " at worldpos " .. tilex .. ", " .. tiley)
	end
end

-------------------------------------------------------------------------------
-- TILE AND CHUNK HANDLING
do
	function world:getchunk(cx, cy)
		return self.chunks[grid.coordinatesToKey(cx, cy)]
	end

	function world:unloadAllChunks()
		self.loaded = {}
	end

	function world:setchunkloaded(cx, cy)
		self.loaded[grid.coordinatesToKey(cx, cy)] = true
	end

	function world:tileUpdate(tilex, tiley)

		local cx, cy, lx, ly = grid.getLocalCoordinates(tilex, tiley)
		local chunk = self:getchunk(cx, cy)
		if chunk then
			chunk.tileupdate[lx][ly] = true
		else
			print("attempted to tileupdate in empty chunk at worldpos " .. tilex .. ", " .. tiley)
		end
	end

	function world:randomTick(wx, wy)
		local tileid = self:getTile(wx, wy)
		local data = tiles:getByID(tileid)

		if data.randomupdate then
			data.randomupdate(self, wx, wy)
		end
	end

	function world:chunkRandomTileUpdates(chunk)
		local wx, wy = chunk.position.x, chunk.position.y
		for i = 1, config.RANDOM_TILE_UPDATES_PER_FRAME do
			local randx = math.random(1, config.CHUNK_SIZE)
			local randy = math.random(1, config.CHUNK_SIZE)
			--print("randomtick", randx, randy)
			self:randomTick((wx * config.CHUNK_SIZE) + randx, (wy * config.CHUNK_SIZE) + randy)
		end
	end

	function world:chunkTileUpdates(chunk)
		local count = 0
		local wx, wy = chunk.position.x, chunk.position.y
		for dx = 1, config.CHUNK_SIZE do
			for dy = 1, config.CHUNK_SIZE do
				-- the tile update
				if count > config.MAX_TILE_UPDATES_PER_FRAME then
					return
				end
				if chunk.tileupdate[dx][dy] == true then
					chunk.tileupdate[dx][dy] = false
					local tile = chunk.tiles[dx][dy]
					if tile > 0 then
						local data = tiles:getByID(tile)

						if data.tileupdate then
							data.tileupdate(self, (wx * config.CHUNK_SIZE) + dx - 1, (wy * config.CHUNK_SIZE) + dy - 1)
							--print("tileupdate", chunkx+dx, chunky+dy, data.name, (data.tileupdate~=nil))
							count = count + 1
						end
					end
				end
			end
		end
	end

	function world:updateChunks(dt)
		-- first, clear the field for all previous chunks
		self:unloadAllChunks()

		-- then set the field for the chunks we need this frame
		-- if a chunk is already generated, then we won't have any extra work
		-- if a chunk is needed, we'll generate it,
		-- if a chunk is no longer needed, discard it
		local fx, fy = self.focuspoint.x, self.focuspoint.y
		-- chunks are generated and kept in memory for an extra # beyond what is rendered.
		-- this is done to enable second-pass structure generation
		-- without having tiles missing on chunk boundaries

		local genradius = config.CHUNK_DRAW_RADIUS + config.CHUNK_SIMULATION_RADIUS + config.CHUNK_BUFFER_RADIUS

		for cx = -genradius, genradius do
			for cy = -genradius, genradius do
				self:setchunkloaded(fx + cx, fy + cy) -- toggles the "load" field for that chunk
			end
		end

		-- cull chunks that are no longer "loaded"
		for key, chunk in pairs(self.chunks) do
			if not self.loaded[key] then
				self.chunks[key] = nil
				self.firstPassComplete[key] = nil
				self.secondPassComplete[key] = nil
				self.fullSimulation[key] = nil
			end
		end
		-- call firstpassgeneration on "loaded" chunks
		-- this is wrapped in a function so as to not attempt
		-- to generate all of the needed chunks in one frame
		-- would result in lag spikes

		-- we don't bother with secondpassgeneration as it's much less intensive
		-- (for now, subject to change later)
		local function chunkloading()
			for key, state in pairs(self.loaded) do
				if state == true then
					if not self.chunks[key] then
						-- generate chunk
						local cx, cy = grid.keyToCoordinates(key)

						-- new chunk defined
						local chunk = {
							position = jutils.point.new(cx, cy),
							tiles = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE), -- 2D array with default 0
							states = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE),
							damage = jutils.table.construct(jutils.table.construct(0, config.CHUNK_SIZE), config.CHUNK_SIZE),
							tileupdate = jutils.table.construct(jutils.table.construct(false, config.CHUNK_SIZE), config.CHUNK_SIZE) -- 2D array with default false
						}

						-- first pass complete
						self.terrainGenerator(chunk)
						self.firstPassComplete[key] = true
						self.chunks[key] = chunk

						return -- end function after generating one chunk per frame
					end
				end
			end
		end
		local function structuregeneration()
			for key, state in pairs(self.firstPassComplete) do
				if not self.secondPassComplete[key] then
					local chunk = self.chunks[key]
					local wx, wy = chunk.position.x, chunk.position.y
					local readyForSecondPass = true

					for cx = -config.CHUNK_BUFFER_RADIUS, config.CHUNK_BUFFER_RADIUS do
						for cy = -config.CHUNK_BUFFER_RADIUS, config.CHUNK_BUFFER_RADIUS do
							if not self.chunks[grid.coordinatesToKey(wx + cx, wy + cy)] then
								readyForSecondPass = false
							end
						end
					end

					if readyForSecondPass then
						for x = 1, config.CHUNK_SIZE do
							for y = 1, config.CHUNK_SIZE do
								self.structureGenerator(self, ((wx) * config.CHUNK_SIZE) + x - 1, ((wy) * config.CHUNK_SIZE) + y - 1)
							end
						end

						self.secondPassComplete[key] = true
						return
					end
				end
			end
		end

		-- encapsulated into their own functions so we can return after doing 1 chunk per frame
		-- ghetto lag prevention LUL
		chunkloading()
		structuregeneration()

		-- do chunk simulations
		local simradius = config.CHUNK_DRAW_RADIUS + config.CHUNK_SIMULATION_RADIUS

		self.fullSimulation = {}
		for cx = -simradius, simradius do
			for cy = -simradius, simradius do
				local key = grid.coordinatesToKey(fx + cx, fy + cy)

				if self.secondPassComplete[key] == true and self.firstPassComplete[key] == true then
					self.fullSimulation[key] = true

					local chunk = self.chunks[key]
					self:chunkRandomTileUpdates(chunk)
					self:chunkTileUpdates(chunk)
				end
			end
		end
	end

	local function getTile(id)
		for name, tiledata in pairs(tiles) do
			--print(name, tiledata.id)
			if tiledata.id == id then
				return tiledata, name
			end
		end
		error("No tile with ID " .. id .. " exists!")
	end

	function world:drawChunks()
		local centerx, centery = self.focuspoint.x, self.focuspoint.y

		for deltax = -config.CHUNK_DRAW_RADIUS, config.CHUNK_DRAW_RADIUS do
			for deltay = -config.CHUNK_DRAW_RADIUS, config.CHUNK_DRAW_RADIUS do
				local chunkx = centerx + deltax
				local chunky = centery + deltay

				local key = grid.coordinatesToKey(chunkx, chunky)

				local chunk = self.chunks[key]

				if chunk and self.fullSimulation[key] then
					local wx = chunkx * config.CHUNK_SIZE
					local wy = chunky * config.CHUNK_SIZE

					local bgcolor = {0.1, 0.3, 0.8} -- sky color
					if wy > config.UNDERGROUND_DEPTH then
						bgcolor = {0.05, 0.05, 0.05}
					end

					love.graphics.setColor(bgcolor)
					love.graphics.rectangle(
						"fill",
						(wx) * config.TILE_SIZE,
						(wy) * config.TILE_SIZE,
						(config.CHUNK_SIZE) * config.TILE_SIZE,
						(config.CHUNK_SIZE) * config.TILE_SIZE
					)

					for x = 1, config.CHUNK_SIZE do
						for y = 1, config.CHUNK_SIZE do
							local tileid = chunk.tiles[x][y]
							local damage = chunk.damage[x][y]
							local state = chunk.states[x][y]
							local light = self.ambientlight

							if tileid > 0 then
								rendering:draw(tileid, state, damage, 1, wx + x - 1, wy + y - 1)
							end
						end
					end
					if config.debug.DRAW_CHUNK_BORDERS then
						love.graphics.setColor(0.5, 0.5, 0.5)
						love.graphics.line(
							(wx) * config.TILE_SIZE,
							(wy) * config.TILE_SIZE,
							((wx) * config.TILE_SIZE) + (config.CHUNK_SIZE * config.TILE_SIZE),
							((wy) * config.TILE_SIZE)
						)
						love.graphics.line(
							(wx) * config.TILE_SIZE,
							(wy) * config.TILE_SIZE,
							((wx) * config.TILE_SIZE),
							((wy) * config.TILE_SIZE) + config.CHUNK_SIZE * config.TILE_SIZE
						)
					end
				end
			end
		end
	end
end
----------------------------------------------------------------------------------
-- ENTITY AND COLLISION HANDLING


function world:updatePhysics(dt)

	if dt > (1/4) then dt = 1/4 end
	for idx, entity in pairs(self.entities) do
		if entity:isA("PhysicalEntity") then
			-- do collision solving
			entity.falling = true

			local function testCollision(tilex, tiley)
				local tileid = self:getTile(tilex, tiley)
				if not (tileid > 0) then return false end
				
				local tiledata = tiles:getByID(tileid)
				if tiledata.collide == false then return false end

				local ex, ey, ew, eh = entity.nextposition.x, entity.nextposition.y, entity.boundingbox.x, entity.boundingbox.y

				local tx, ty, tw, th = (tilex*config.TILE_SIZE)+(config.TILE_SIZE/2), (tiley*config.TILE_SIZE)+(config.TILE_SIZE/2), config.TILE_SIZE/2, config.TILE_SIZE/2

				local sx, sy = collision:test(ex, ey, ew, eh, tx, ty, tw, th)

				if not (sx and sy) then return false end

				local separation = jutils.point.new(sx, sy)

				local normalx, normaly = collision:solve(sx, sy, entity.velocity.x, entity.velocity.y)

				if tiledata.customCollision then
					tiledata.customCollision(entity, separation, jutils.point.new(normalx, normaly))
				elseif tiledata.solid then-- default collision solver
					if normalx ~= nil and normaly ~= nil then
						entity.nextposition = entity.nextposition + separation
						-- collides with top of tile
						if normaly == -1 then
							entity.velocity.y = 0
							entity.falling = false
						end

						if normaly == 1 then
							entity.velocity.y = -(entity.velocity.y * 0.3)
						end
					end
					entity:collisionCallback(tileid, jutils.point.new(tilex, tiley), separation, jutils.point.new(normalx, normaly))
				end
				
			end

			if entity.noclip == false then
				local entityTileX, entityTileY = grid.pixelToTileXY(entity.nextposition.x, entity.nextposition.y)
				local extx, exty = grid.pixelToTileXY(entity.boundingbox.x, entity.boundingbox.y)
				extx = extx + 1
				exty = exty + 1

				-- tile extents around the player
				for x = -extx, extx do
					for y = -exty, exty do
						testCollision(entityTileX+x, entityTileY+y)
					end
				end
			end

			entity:updatePhysics(dt)

		end
	end
end

function world:updateEntities(dt)
	for idx, entity in pairs(self.entities) do
		entity:update(dt)

		if entity.dead then
			entity:dies()
			entity.world = nil
			self.entities[idx] = nil
		end

		local x, y = self.focuspoint.x*config.CHUNK_SIZE*config.TILE_SIZE, self.focuspoint.y*config.CHUNK_SIZE*config.TILE_SIZE

		local radius = config.CHUNK_DRAW_RADIUS+config.CHUNK_SIMULATION_RADIUS+config.CHUNK_BUFFER_RADIUS

		radius = (radius * config.CHUNK_SIZE) * config.TILE_SIZE



		if not entity:isA("Player") then
			if entity.position:distance(jutils.point.new(x, y)) > radius then
				entity.unloadtimer = entity.unloadtimer + dt
			else
				entity.unloadtimer = 0
			end

			if entity.unloadtimer > 1 then
				entity.dead = true
			end
		end
	end
end

function world:update(dt)
	self:updateChunks(dt)
	self:updatePhysics(dt)
	self:updateEntities(dt)
	
end

function world:drawEntities()
	for _, entity in pairs(self.entities) do
		entity:draw()
	end
end

function world:draw()
	self.dx, self.dy = love.graphics.inverseTransformPoint(love.mouse.getX(), love.mouse.getY())

	self:drawChunks()
	self:drawEntities()
end

return world
